import os
from threading import Thread

import click

from release_exporter.exceptions import UnknownRepo
from release_exporter.formatter import github
from release_exporter.formatter import gitlab
from release_exporter.utils import get_repo_url_info, check_version, Init
from release_exporter.version import __version__


def thread_caller():
    th = Thread(target=check_version)
    th.start()


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(__version__)
    ctx.exit()


@click.group()
@click.option("--token", help="Token number if its a private repository.", default=None)
@click.option(
    "--url",
    help="URL of your repository. This is optional if your current directory has .git folder with remote url.",
    default=None,
)
@click.option("--location", help="Local location of your repository.", default=os.getcwd())
@click.option("--version", is_flag=True, callback=print_version, expose_value=False, is_eager=True)
@click.option("--universal", help="Create a global settings file. Defaults to True.", default=True)
@click.pass_context
def cli(ctx, token, url, location, universal):
    ctx.obj["token"] = token
    ctx.obj["repo_url"] = url
    ctx.obj["location"] = location
    ctx.obj["universal"] = universal


@cli.command(help="Creates .rex file.")
def init(ctx):
    if ctx.obj["universal"]:
        Init().config()
    else:
        Init(os.getcwd()).config()


@cli.command(help="Creates markdown file.")
@click.pass_context
def markdown(ctx):
    if os.name == "nt":
        ctx.obj["location"] = rf"{ctx.obj['location']}"

    if "github" in get_repo_url_info(location=ctx.obj["location"], repo_url=ctx.obj["repo_url"]).resource:
        print("GitHub detected. \n")

        github(
            force=True,
            token=ctx.obj["token"],
            location=ctx.obj["location"],
            repo_url=ctx.obj["repo_url"],
            file_type="markdown",
        ).write()

    elif "gitlab" in get_repo_url_info(location=ctx.obj["location"], repo_url=ctx.obj["repo_url"]).resource:
        print("GitLab detected. \n")

        gitlab(
            force=True,
            token=ctx.obj["token"],
            location=ctx.obj["location"],
            repo_url=ctx.obj["repo_url"],
            file_type="markdown",
        ).write()

    else:
        raise UnknownRepo("Sorry, couldn't find the repository. Try giving the repository URL by adding --url flag.")


@cli.command(help="Creates JSON file.")
@click.pass_context
def json(ctx):
    if os.name == "nt":
        ctx.obj["location"] = rf"{ctx.obj['location']}"

    if "github" in get_repo_url_info(location=ctx.obj["location"], repo_url=ctx.obj["repo_url"]).resource:
        click.echo("GitHub detected. \n")

        github(
            force=True,
            token=ctx.obj["token"],
            location=ctx.obj["location"],
            repo_url=ctx.obj["repo_url"],
            file_type="json",
        ).write()

    elif "gitlab" in get_repo_url_info(location=ctx.obj["location"], repo_url=ctx.obj["repo_url"]).resource:
        click.echo("GitLab detected. \n")

        gitlab(
            force=True,
            token=ctx.obj["token"],
            location=ctx.obj["location"],
            repo_url=ctx.obj["repo_url"],
            file_type="json",
        ).write()

    else:
        raise UnknownRepo("Sorry, couldn't find the repository. Trying giving the repository URL by adding --url flag.")


@cli.command(help="Creates reStructuredText file. Coming soon.")
@click.pass_context
def rst(ctx):
    # if os.name == 'nt':
    #     ctx.obj['location'] = r'{}'.format(ctx.obj['location'])
    #
    # if "github" in get_repo_url_info(location=ctx.obj['location'], repo_url=ctx.obj['repo_url']).resource:
    #     click.echo('GitHub detected. \n')
    #
    #     github(force=True, token=ctx.obj['token'], location=ctx.obj['location'], repo_url=ctx.obj['repo_url'],
    #            file_type='rst').write()
    #
    # elif "gitlab" in get_repo_url_info(location=ctx.obj['location'], repo_url=ctx.obj['repo_url']).resource:
    #     click.echo('GitLab detected. \n')
    #
    #     gitlab(force=True, token=ctx.obj['token'], location=ctx.obj['location'], repo_url=ctx.obj['repo_url'],
    #            file_type='rst').write()
    #
    # else:
    #     raise UnknownRepo("Sorry, couldn't find the repository. Trying giving the repository URL by adding --url flag.")
    raise NotImplementedError("Coming soon.")


@cli.command("all", help="Creates change log for all formats.")
@click.pass_context
def all_format(ctx):
    if os.name == "nt":
        ctx.obj["location"] = rf"{ctx.obj['location']}"

    if "github" in get_repo_url_info(location=ctx.obj["location"], repo_url=ctx.obj["repo_url"]).resource:
        # Creates for GitHub
        print("Creating change logs for GitHub.")
        github(
            force=True,
            token=ctx.obj["token"],
            location=ctx.obj["location"],
            repo_url=ctx.obj["repo_url"],
            file_type="json",
        ).write()
        github(
            force=True,
            token=ctx.obj["token"],
            location=ctx.obj["location"],
            repo_url=ctx.obj["repo_url"],
            file_type="markdown",
        ).write()
        # github(force=True, token=ctx.obj['token'], location=ctx.obj['location'], repo_url=ctx.obj['repo_url'],
        #        file_type='rst').write()
    else:
        # Creates for GitLab
        print("Creating change logs for GitLab.")
        gitlab(
            force=True,
            token=ctx.obj["token"],
            location=ctx.obj["location"],
            repo_url=ctx.obj["repo_url"],
            file_type="json",
        ).write()
        gitlab(
            force=True,
            token=ctx.obj["token"],
            location=ctx.obj["location"],
            repo_url=ctx.obj["repo_url"],
            file_type="markdown",
        ).write()
        # gitlab(force=True, token=ctx.obj['token'], location=ctx.obj['location'], repo_url=ctx.obj['repo_url'],
        #        file_type='rst').write()


def main():
    thread_caller()
    cli(obj={})



import os
import pathlib

from .exceptions import FileExists

FILE_TYPE = {"markdown": ".md", "json": ".json", "rest": ".rst"}


class FormatBase:
    """
    Base formatter for logs.
    """

    def __init__(
        self,
        token=None,
        request_header=None,
        api_url=None,
        info=None,
        repo_url=None,
        location=os.getcwd(),
        file_name="CHANGELOG",
        force=False,
        file_type="markdown",
    ):
        self.location = location
        self.file_name = file_name
        self.force = force
        self.file_type = file_type
        self.all_content = []
        self.token = token
        self.request_header = request_header
        self.api_url = api_url
        self.info = info
        self.repo_url = repo_url

        self.tag_name = None
        self.date = None
        self.description = None
        self.compare_url = None
        self.total_number_tags = None
        self.iter_count = None
        self.list_descriptions = []

        if not force:
            if pathlib.Path(location + os.sep + file_name + self._get_file_ext()).is_file():
                raise FileExists(
                    f"{file_name} {self._get_file_ext()} already exists at {location}. Use --force to override or force=True.'"
                )

    def _get_file_ext(self):
        try:
            return FILE_TYPE[self.file_type]
        except KeyError as e:
            print(e)

    def write(self):
        raise NotImplementedError("Coming soon.")

    @classmethod
    def _header(cls):
        return """\
# Changelog

All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](http://keepachangelog.com/en/1.0.0/)
and this project adheres to [Semantic Versioning](http://semver.org/spec/v2.0.0.html).

## [Unreleased]

"""

    def _body(self):
        if self.iter_count < self.total_number_tags - 1:
            return f"""\
    
## [{self.tag_name}] - {self.date}

{self.description}
"""
        else:
            return f"""\

## {self.tag_name} - {self.date}

{self.description}

"""

    def _footer(self):
        return """\
[{tag_name}]: {url}
""".format(tag_name=self.tag_name, url=self.repo_url)

    def _converter(self):
        pass

    def _dict_repo_template(self):
        repo = {
            "repositoryName": self.info.name,
            "provider": self.info.resource,
            "owner": self.info.owner,
            "repoUrl": "https://" + self.info.resource + "/" + self.info.owner + "/" + self.info.name,
            "totalTags": self.total_number_tags,
            "data": self.list_descriptions,
        }

        return repo

    @classmethod
    def _dict_data_template(cls, tag_name=None, description=None, date=None, compare_url=None):
        data = {"tagName": tag_name, "description": description, "createdAt": date, "compareUrl": compare_url}

        return data

    def _total_number_releases(self):
        pass

    def releases(self):
        pass

    @classmethod
    def _header_rst(cls):
        return """\
Changelog
=========

All notable changes to this project will be documented in this file.

The format is based on `Keep a Changelog <https://keepachangelog.com/en/1.0.0/>`__
and this project adheres to `Semantic Versioning <https://semver.org/spec/v2.0.0.html>`__.

Unreleased_
-----------

"""

    def _body_rst(self):
        if self.iter_count < self.total_number_tags - 1:
            return f"""\

{self.tag_name}_ - {self.date}
~~~~~~~~~~~~~~~~~~~~~~~~~~

{self.description}
"""
        else:
            return f"""\

{self.tag_name} - {self.date}
~~~~~~~~~~~~~~~~~~~~~~~~~~

{self.description}

"""

    def _footer_rst(self):
        return f"""\
.. _{self.tag_name}: {self.repo_url}
"""



class FileExists(Exception):
    """
    This exception is raised when a file exists.
    """

    def __init__(self, message, errors=None):
        super(FileExists, self).__init__(message)

        self.errors = errors


class ParserError(Exception):
    """
    This exception is raised Git config file does not exist.
    """

    def __init__(self, message, errors=None):
        super(ParserError, self).__init__(message)

        self.errors = errors


class UnknownRepo(Exception):
    """
    This exception is raised Git config file does not exist.
    """

    def __init__(self, message, errors=None):
        super(UnknownRepo, self).__init__(message)

        self.errors = errors


class InvalidToken(Exception):
    """
    This exception is raised when an invalid token is given or the token is empty (None).
    """

    def __init__(self, message, errors=None):
        super(InvalidToken, self).__init__(message)

        self.errors = errors



import json

from release_exporter.request import GitHubRequest, GitLabRequest
from release_exporter.utils import date_convert, description


class GitHubFormat(GitHubRequest):
    """
    Changelog of GitHub.
    """

    def __init__(self, *args, **kwargs):
        super(GitHubFormat, self).__init__(*args, **kwargs)

        self.compare_url = "https://" + self.info.resource + "/" + self.info.owner + "/" + self.info.name + "/compare/"

    def write(self):
        """
        Write files according to the file type chosen. This method support ``markdown``, ``json`` and ``rst``.

        :raises: ValueError
        """

        if self.file_type == "markdown":
            with open("CHANGELOG.md", "w") as md_file:
                md_file.writelines(self._converter())

            print("\n" + "Markdown file created!")

        elif self.file_type == "json":
            self._converter()

            with open(self.file_name + "." + self.file_type, "w") as json_file:
                json.dump(self._dict_repo_template(), json_file, indent=4)

            print("\n" + "JSON file created!")
        elif self.file_type == "rst":
            raise NotImplementedError("Coming soon")
        else:
            raise ValueError("Unknown file format specified.")

    def _converter(self):
        """
        A tuple of formatted tag name, description, created at and the compare links.

        :return:
        :type: tuple -- A tuple of list.
        """

        temp = self.releases()["releases"]
        temp = [i for i in temp if i["node"]["tag"]]
        temp_l = []

        self.total_number_tags = sum(1 for k in temp if k["node"]["tag"]["name"])

        print(description(provider=self.info.resource, repo_name=self.info.name, tags_number=self.total_number_tags))

        if self.file_type == "markdown":
            self.all_content.append(self._header())

            for count, edge in enumerate(temp):
                self.iter_count = count
                temp_l.append(edge["node"]["tag"]["name"])
                self.tag_name = edge["node"]["tag"]["name"]
                self.description = edge["node"]["description"].replace("\r\n", "\n")
                self.date = date_convert(edge["node"]["createdAt"])
                self.all_content.append(self._body())

            pair = list(["{}...{}".format(a, b) for a, b in zip(temp_l, ["HEAD"] + temp_l[:-1])])

            for count, tags in enumerate(pair):
                if count < 1:
                    self.all_content.append("[Unreleased]: " + self.compare_url + tags + "\n")
                else:
                    self.all_content.append("[" + tags.split("...")[1] + "]: " + self.compare_url + tags + "\n")

            return tuple(self.all_content)

        elif self.file_type == "json":
            temp_l2 = []
            tag_comp_url_temp = []
            self.list_descriptions.insert(0, self._dict_data_template(tag_name="Unreleased", description="", date=""))

            for count, edge in enumerate(temp):
                self.iter_count = count
                temp_l2.append(edge["node"]["tag"]["name"])

                self.list_descriptions.append(
                    self._dict_data_template(
                        tag_name=edge["node"]["tag"]["name"],
                        description=edge["node"]["description"].replace("\r\n", "\n"),
                        date=date_convert(edge["node"]["createdAt"]),
                    )
                )

            pair = list(["{}...{}".format(a, b) for a, b in zip(temp_l2, ["HEAD"] + temp_l2[:-1])])

            for tags in pair:
                tag_comp_url_temp.append(self.compare_url + tags)

            for count, urls in enumerate(tag_comp_url_temp):
                self.list_descriptions[count]["compareUrl"] = urls

        elif self.file_type == "rst":
            raise NotImplementedError("Coming soon")


github = GitHubFormat


class GitLabFormat(GitLabRequest):
    """
    Changelog of GitLab.
    """

    def __init__(self, *args, **kwargs):
        super(GitLabFormat, self).__init__(*args, **kwargs)

        self.compare_url = "https://" + self.info.resource + "/" + self.info.owner + "/" + self.info.name + "/compare/"

    def write(self):
        """
        Write files according to the file type chosen. This method support ``markdown``, ``json`` and ``rst``.

        :raises: ValueError
        """

        if self.file_type == "markdown":
            with open("CHANGELOG.md", "w") as md_file:
                md_file.writelines(self._converter())

            print("\n" + "Markdown file created!")

        elif self.file_type == "json":
            self._converter()

            with open(self.file_name + "." + self.file_type, "w") as json_file:
                json.dump(self._dict_repo_template(), json_file, indent=4)

            print("\n" + "JSON file created!")
        elif self.file_type == "rst":
            raise NotImplementedError("Coming soon")
        else:
            raise ValueError("Unknown file format specified.")

    def _converter(self):
        """
        A tuple of formatted tag name, description, created at and the compare links.

        :return:
        :type: tuple -- A tuple of list.
        """

        temp = self.releases()
        temp_l = []

        self.total_number_tags = sum(1 for k in temp if k["name"])

        print(description(provider=self.info.resource, repo_name=self.info.name, tags_number=self.total_number_tags))

        if self.file_type == "markdown":
            self.all_content.append(self._header())

            for count, content in enumerate(temp):
                self.iter_count = count
                temp_l.append(content["name"])
                self.tag_name = content["name"]
                self.description = content["release"]["description"].replace("\r\n", "\n")
                self.date = date_convert(content["commit"]["created_at"])
                self.all_content.append(self._body())

            pair = list(["{}...{}".format(a, b) for a, b in zip(temp_l, ["HEAD"] + temp_l[:-1])])

            self.all_content.append("\n")

            for count, tags in enumerate(pair):
                if count < 1:
                    self.all_content.append("[Unreleased]: " + self.compare_url + tags + "\n")
                else:
                    self.all_content.append("[" + tags.split("...")[1] + "]: " + self.compare_url + tags + "\n")

            return tuple(self.all_content)

        elif self.file_type == "json":
            temp_l2 = []
            tag_comp_url_temp = []
            self.list_descriptions.insert(0, self._dict_data_template(tag_name="Unreleased", description="", date=""))

            for count, content in enumerate(temp):
                self.iter_count = count
                temp_l2.append(content["name"])

                self.list_descriptions.append(
                    self._dict_data_template(
                        tag_name=content["name"],
                        description=content["release"]["description"].replace("\r\n", "\n"),
                        date=date_convert(content["commit"]["created_at"]),
                    )
                )

            pair = list(["{}...{}".format(a, b) for a, b in zip(temp_l2, ["HEAD"] + temp_l2[:-1])])

            for tags in pair:
                tag_comp_url_temp.append(self.compare_url + tags)

            for count, urls in enumerate(tag_comp_url_temp):
                self.list_descriptions[count]["compareUrl"] = urls

        elif self.file_type == "rst":
            raise NotImplementedError("Coming soon")


gitlab = GitLabFormat



import json

import requests

from release_exporter.base import FormatBase
from release_exporter.exceptions import InvalidToken
from release_exporter.utils import get_repo_url_info, multi_key_gitlab


class GitHubRequest(FormatBase):
    """
    GitHub request base.
    """

    def __init__(self, *args, **kwargs):
        super(GitHubRequest, self).__init__(*args, **kwargs)
        self.request_headers = {"Authorization": "token %s" % self.token}
        self.api_url = "https://api.github.com/graphql"

        if self.token is None:
            raise InvalidToken(
                "Oops! GitHub requires you to generate a private token to get the details. See "
                "https://help.github.com/articles/creating-a-personal-access-token-for-the-command-line/ "
                "for more information."
            )

        if self.repo_url is not None:
            self.info = get_repo_url_info(self.location, repo_url=self.repo_url)
        else:
            self.info = get_repo_url_info(self.location)

    def _total_number_releases(self):
        """
        Queries total number of releases.

        :returns: Number of releases.
        :rtype: int
        """
        _json = {
            "query": f"""
            {{
              repository(owner: \"{self.info.owner}\", name: \"{self.info.name}\") {{
                releases {{
                  totalCount
                }}
              }}
            }}
        """
        }

        r = requests.post(url=self.api_url, json=_json, headers=self.request_headers)

        try:
            response_data = r.json()
        except json.JSONDecodeError:
            raise requests.HTTPError(f"Invalid JSON response: {r.text}")

        if r.status_code >= 400:
            raise requests.HTTPError(f"Request failed with status code {r.status_code}: {r.text}")

        if "errors" in response_data:
            error_messages = [error.get("message", "Unknown error") for error in response_data["errors"]]
            raise ValueError(f"GraphQL query failed with errors: {', '.join(error_messages)}")

        try:
            return int(response_data["data"]["repository"]["releases"]["totalCount"])
        except (TypeError, KeyError):
            raise ValueError("Repository data not found in response")

    def releases(self) -> dict[str, list]:
        """
        A JSON object with name of the repository, tag name, description and the created date and time.

        :returns: A dict object
        :rtype: dict
        """
        result = {"releases": []}
        after_cursor = None
        total = self._total_number_releases()

        while total > 0:
            after_query = f', after: "{after_cursor}"' if after_cursor else ""
            _json = {
                "query": f"""
                        query {{
                          repository(owner: \"{self.info.owner}\", name: \"{self.info.name}\") {{
                            releases(
                              first: 100
                              {after_query}
                              orderBy: {{field: CREATED_AT, direction: DESC}}
                            ) {{
                              pageInfo {{
                                hasNextPage
                                endCursor
                              }}
                              edges {{
                                node {{
                                  name
                                  tag {{
                                    name
                                  }}
                                  description
                                  createdAt
                                }}
                              }}
                            }}
                          }}
                        }}
                        """
            }

            r = requests.post(url=self.api_url, json=_json, headers=self.request_headers)

            try:
                response_data = r.json()
            except json.JSONDecodeError:
                raise requests.HTTPError(f"Invalid JSON response: {r.text}")

            new_releases = response_data["data"]["repository"]["releases"]["edges"]
            result["releases"].extend(new_releases)

            # Check if there are more pages
            page_info = response_data["data"]["repository"]["releases"]["pageInfo"]
            if page_info["hasNextPage"]:
                after_cursor = page_info["endCursor"]
                total -= 100  # Assuming each page returns 100 results
            else:
                break

        return result


class GitLabRequest(FormatBase):
    """
    GitLab request base.
    """

    def __init__(self, *args, **kwargs):
        super(GitLabRequest, self).__init__(*args, **kwargs)
        self.request_headers = {"Private-Token": "%s" % self.token}
        self.api_url = "https://gitlab.com/api/v4/"

        if self.token is None:
            raise InvalidToken(
                "Oops! GitLab requires you to generate a private token to get the details. See "
                "https://docs.gitlab.com/ce/user/profile/personal_access_tokens.html "
                "for more information."
            )

        if self.repo_url is not None:
            self.info = get_repo_url_info(self.location, repo_url=self.repo_url)
        else:
            self.info = get_repo_url_info(self.location)

    def _repo_id(self):
        """
        Searches and returns the repository ID based on the repository name. If the repository is not found then a table
        of repository is shown so that the user can manually enter the ID of their repository.

        :returns: Repository ID
        :return: id_number
        :rtype: int

        """
        url = self.api_url + f"projects?search={self.info.name}"

        r = requests.get(url=url, headers=self.request_headers)
        id_number = json.loads(r.text)

        if len(id_number) > 1:
            print(
                "The search resulted in more that one repository. Please check your repository name and type in it's ID"
            )
            print("ID - Repository Name - Username")

            for content in id_number:
                print(f"{content['id']} - {content['name']} - {multi_key_gitlab(content)}")

            id_number = input("ID > ")
            return id_number

        try:
            return id_number[0]["id"]
        except KeyError:
            raise KeyError("Wrong credentials given. Please check if you have the correct token.")

    def releases(self):
        """
        A JSON object containing name of the repository, tag name, description and the created date and time

        :returns: A dict object
        :rtype: dict
        """
        url = f"https://gitlab.com/api/v4/projects/{self._repo_id()}/repository/tags"

        r = requests.get(url=url, headers=self.request_headers)
        return json.loads(r.text)



import configparser
import datetime
import json
import os
import warnings
from packaging.version import Version
from pathlib import Path

import dateutil.parser
import requests
from giturlparse import parse
from requests.exceptions import ConnectionError
from tabulate import tabulate

from release_exporter import version
from release_exporter.exceptions import ParserError

CONFIG_FILE_NAME = ".rex"
VERSION_API_URL = "https://pypi.org/pypi/release-exporter/json"


def get_repo_url_info(location=os.getcwd(), repo_url=None):
    """
    Returns the parsed URL.

    Parameters
    ----------
    location: str
        Absolute location of the current directory.
    repo_url: str
        URL of the repository.

    Returns
    -------
    parse: giturlparse.parser.Parsed
        A named tuple.

    """
    try:
        if repo_url is None:
            config = configparser.ConfigParser()
            config.read(location + os.sep + ".git" + os.sep + "config")
            if 'remote "origin"' in config.sections():
                return parse(config['remote "origin"']["url"])
            else:
                raise ParserError("Git config file does not exist please provide the repository url by using --url.")
        else:
            return parse(repo_url + ".git")
    except configparser.DuplicateSectionError:
        raise configparser.DuplicateSectionError(
            "There seems to be a duplicate section in your config. Try giving the repository URL by using --url."
        )


def date_convert(date):
    """
    Converts ISO8601 date and time and returns only the date.

    Parameters
    ----------
    date: str
        datetime string.

    Returns
    -------
    date: str
        Date as Y-m-d format..

    """
    try:
        date = datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%Sz")
        date = date.strftime("%Y-%m-%d")
    except ValueError:
        date = dateutil.parser.parse(date).date().strftime("%Y-%m-%d")
    return date


def multi_key_gitlab(value):
    """
    Returns the username, if an exception occurs None is returned.

    Parameters
    ----------
    value: dict
        A dictionary of GitLab.

    Returns
    -------
    value: str or None
        Username or none.
    """

    try:
        return value["owner"]["username"]
    except (KeyError, TypeError):
        return None


def description(provider=None, repo_name=None, tags_number=None):
    """
    Description generator.

    Parameters
    ----------
    provider: str
        Name of the Git host.
    repo_name: str
        Repository name.
    tags_number: str or int
        Number of tags.

    Return
    ------
    tabulate: str
        A tabulated structure of the input.
    """
    table = [["Provider", provider], ["Repository Name", repo_name], ["Number of Tags", tags_number]]

    return tabulate(table, tablefmt="grid")


class Init:
    """
    Creates ``.rex`` and ``CHANGELOG`` file depending on the format given, if no format if given ``CHANGELOG.md`` is
    created.

    >>> rex init
    Created .rex config file at <user home path>
    """

    def __init__(self, location=Path.home()):
        self.location = location

        self.config_file_path = Path(os.path.join(self.location, CONFIG_FILE_NAME))

    @property
    def config(self):
        if self.config_file_path.is_file():
            config_file = configparser.ConfigParser()
            config_file.read(self.config_file_path)
            return self.config_file_path, config_file
        else:
            self._init_config()

        return self.config_file_path

    def _init_config(self):
        config = configparser.ConfigParser()

        config["DEFAULT"] = {
            "Version": version.__version__,
            "CheckCycle": 10,
            "GitHubKey": "",
            "GitLabKey": "",
            "LastChecked": datetime.date.today(),
        }

        with open(self.config_file_path, "w") as configfile:
            config.write(configfile)

    @config.setter
    def config(self, config_dict):
        assert isinstance(config_dict, dict)

        self.config_file_path, config_file = self.config

        exists, config, dict_key = self._check_config_property(config_dict)

        if not exists:
            if config_dict[dict_key]["Key"]:
                value = (
                    config["DEFAULT"]["githubkey"]
                    if config_dict[dict_key]["Provider"] == "github"
                    else config["DEFAULT"]["gitlabkey"]
                )
                config_dict[dict_key]["Key"] = value

            config.read_dict(config_dict)
            with open(self.config_file_path, "w") as configfile:
                config.write(configfile)

    def _check_config_property(self, config_dict):
        config = configparser.ConfigParser()
        config.read(self.config_file_path)

        dict_key = next(iter(config_dict.keys()))
        exists = dict_key in config.sections()

        return exists, config, dict_key


def check_version():
    """
    Checks for latest update.

    :return: version
    :rtype: str
    :raises: ConnectionError
    """
    from colorama import init, Fore

    init(autoreset=True)
    try:
        r = requests.get(VERSION_API_URL)

        versions = list(json.loads(r.text)["releases"].keys())

        if len(versions) > 1:
            versions = sorted(versions, key=Version)

            if versions[-1] != version.__version__:
                table = [
                    ["New version " + versions[-1] + " available. To update type in"],
                    ["pip install -U release-exporter"],
                ]
                print(Fore.RED + tabulate(table, stralign="center"))
        else:
            pass

    except ConnectionError:
        pass


# Taken from NumPy
def _set_function_name(func, name):
    func.__name__ = name
    return func


class _Deprecate(object):
    """
    Decorator class to deprecate old functions.
    Refer to `deprecate` for details.
    """

    def __init__(self, old_name=None, new_name=None, message=None):
        self.old_name = old_name
        self.new_name = new_name
        self.message = message

    def __call__(self, func, *args, **kwargs):
        """
        Decorator call.  Refer to ``decorate``.
        """
        old_name = self.old_name
        new_name = self.new_name
        message = self.message

        if old_name is None:
            try:
                old_name = func.__name__
            except AttributeError:
                old_name = func.__name__
        if new_name is None:
            depdoc = "`%s` is deprecated!" % old_name
        else:
            depdoc = "`%s` is deprecated, use `%s` instead!" % (old_name, new_name)

        if message is not None:
            depdoc += "\n" + message

        def newfunc(*args, **kwds):
            """``arrayrange`` is deprecated, use `arange` instead!"""
            warnings.warn(depdoc, DeprecationWarning, stacklevel=2)
            return func(*args, **kwds)

        newfunc = _set_function_name(newfunc, old_name)
        doc = func.__doc__
        if doc is None:
            doc = depdoc
        else:
            doc = "\n\n".join([depdoc, doc])
        newfunc.__doc__ = doc
        try:
            d = func.__dict__
        except AttributeError:
            pass
        else:
            newfunc.__dict__.update(d)
        return newfunc


def deprecate(*args, **kwargs):
    """
    Deprecation warning. It can be used as an decorator or as a method.

    :param message: Warning message.
    :param newname: New method name.
    :param oldname: Old method name.
    :return: object.
    """
    if args:
        fn = args[0]
        args = args[1:]

        if "newname" in kwargs:
            kwargs["new_name"] = kwargs.pop("newname")
        if "oldname" in kwargs:
            kwargs["old_name"] = kwargs.pop("oldname")

        return _Deprecate(*args, **kwargs)(fn)
    else:
        return _Deprecate(*args, **kwargs)


deprecate_with_doc = lambda msg: _Deprecate(message=msg)

if __name__ == "__main__":
    a = Init()
    # print(a.config)
    # a.config = {
    #     'github:release-exporter': {
    #         'Key': True,
    #         'URL': 'https://',
    #         'RepoName': 'release-exporter',
    #         'Provider': 'github'
    #     }
    # }
    print(a.config)
